/**
 * @file RobotComponent.h
 *
 * @author Laurent Saint-Marcel
 *
 * Class de base dont herite tous les composants du robot qui doivent etre 
 * reset apres un match. Elle definie des fonctions virtuelles utilisees 
 * pendant la procedure d'initialisation: reset, isInitialized
 */

#ifndef __ROBOT_COMPONENT_H__
#define __ROBOT_COMPONENT_H__

#include "robotBase.h"
#include <list>
#include <assert.h>

class RobotComponent;

typedef std::list<RobotComponent* > RobotComponentList;

// ============================================================================
// ==========================  class RobotComponent   =========================
// ============================================================================

/** 
 * @class RobotComponent
 * Class de base dont herite tous les composants du robot: inputs, move, uart\n
 * Elle definie des fonctions virtuelles de communes
 */
class RobotComponent : public RobotBase
{ 
 public:
    RobotComponent(const char* name, ClassId id, const RobotConfig* config);
    virtual ~RobotComponent();

    /** @brief Renvoie la liste de tous les composants crs par l'application
     */
    static RobotComponentList* RobotComponent::getComponentsList();

    /** @brief Redemarre le composant dans son etat initial et vide les 
               buffers... */
    virtual bool reset();

    /** @brief Cette fonction lance une serie de tests sur le composants pour 
               savoir si le code n'a pas ete boggue */
    virtual bool validate()=0;

    /** @brief Cette fonction retourne TRUE si le composant a ete initialise */
    bool         isInitialized() const;


 private:
    bool operator < (RobotComponent const* r1);

 protected:
    // members
    bool         init_;

 private:
    static RobotComponentList* robotComponentList_;
};

// ============================================================================
// =============================  INLINE FUNCTIONS  ===========================
// ============================================================================

inline bool RobotComponent::operator < (RobotComponent const* r1) 
{
  return reinterpret_cast<int>(this) < reinterpret_cast<int>(r1);
}

// ----------------------------------------------------------------------------
// RobotComponent::RobotComponent
// ----------------------------------------------------------------------------
inline RobotComponent::RobotComponent(const char*        name,
				      ClassId            classId,
				      const RobotConfig* config) :
  RobotBase(name, classId, config), init_(false)
{
  assert(config != NULL);
  if (!robotComponentList_) robotComponentList_ = new RobotComponentList();
  robotComponentList_->push_back(this);
}

// ----------------------------------------------------------------------------
// RobotComponent::~RobotComponent
// ----------------------------------------------------------------------------
inline RobotComponent::~RobotComponent()
{
  RobotComponentList::iterator it=robotComponentList_->begin();
  for(;it!= robotComponentList_->end();it++) {
    if ((*it)==this) {
      robotComponentList_->erase(it);
      break;
    }
  }
}

// ----------------------------------------------------------------------------
// RobotComponent::getComponentsList
// ----------------------------------------------------------------------------
inline RobotComponentList* RobotComponent::getComponentsList() 
{
  return robotComponentList_;
}

// ----------------------------------------------------------------------------
// RobotComponent::isInitialized
// ----------------------------------------------------------------------------
inline bool RobotComponent::isInitialized() const
{
    return init_;
}

// ----------------------------------------------------------------------------
// RobotComponent::reset
// ----------------------------------------------------------------------------
inline bool RobotComponent::reset()
{
    init_=false;
    return isInitialized();
}

#endif /* __ROBOT_COMPONENT_H__ */
