/**
 * @file viewer3D.cpp
 *
 * @author Laurent Saint-Marcel
 *
 * Systeme d'affichage du terrain, du robot et des donnees dans plusieurs
 * fenetres avec OPENGL
 */

#ifndef __VIEWER_3D_H__
#define __VIEWER_3D_H__

#include "mthread.h"
#include "image.h"

#include <GL/gl.h>

#define Viewer3D Viewer3DCL::instance()

/*
#define WHITE_BORDER     0.9f, 0.9f,0.9f
#define YELLOW_GRID_1    0.53f,0.47f,0.24f
#define YELLOW_GRID_2    0.56f,0.596f,0.46f
#define BLUE_BORDER      0.17f,0.28f,0.79f
#define BLUE_BORDER_2    0.0f,0.2f,0.5f
#define GREEN_COCONUT    0.12,0.23f,0.03f
#define RED_BALL         0.46f,0.16f,0.17f
*/

#define WHITE_BORDER     0.9f, 0.9f,0.9f
#define YELLOW_GRID_1    0.73f,0.67f,0.20f
#define YELLOW_GRID_2    0.76f,0.86f,0.40f
#define BLUE_BORDER      0.27f,0.48f,0.79f
#define BLUE_BORDER_2    0.0f,0.2f,0.5f
#define GREEN_COCONUT    0.22,0.53f,0.03f
#define RED_BALL         0.96f,0.16f,0.07f
#define ENEMY_RP         1,1,0.0f
#define ENEMY_RS         1,1,0.0f
#define ROBOT_RS         0,0.2f,1

#define VIEWER_MAX_ROBOT_NBR 4

static const double SCALE3D=0.1;

class Texture;

typedef struct ViewerColorST {
  float r;
  float g;
  float b;
  float a;
  ViewerColor(double R=1, double G=1, double B=1, double A=1) {
    r=(float)R;
    g=(float)G;
    b=(float)B;
    a=(float)A;
  }
  void setGL() {
    glColor4f(r,g,b,a);
  }
} ViewerColorST;

typedef enum ViewerScreenEN {
  VIEWER_SCREEN_MAP,
  VIEWER_SCREEN_3D,
  VIEWER_SCREEN_PLAYER,
  
  VIEWER_SCREEN_NBR
} ViewerScreenEN;

typedef enum ViewerRobotModel {
    VIEWER_ROBOT_MODEL1,
    VIEWER_ROBOT_2,

    VIEWER_ROBOT_NBR
} ViewerRobotModel

typedef struct ViewerMovieDataST {
  bool  run;
  int   count;
  char  filename[255];
  Image img;

  ViewerMovieDataST(): run(false), count(0), img(){}
} ViewerMovieDataST;

typedef enum ViewerButton {
    VIEWER_ROBOT_MODEL1,
    VIEWER_ROBOT_2,

    VIEWER_ROBOT_NBR
} ViewerRobotModel

/** 
 * @class Viewer3DCL
 * @brief Class qui gere toute l'interface avec le systeme d'affichage du 
 * simulateur et du logViewer
 * Vue du terrain en 3D + carte du terrain + donnees
 */
class Viewer3DCL
{
 public:
	virtual ~Viewer3DCL();
	static Viewer3DCL* instance();

        // ---------------------------------------------------------------
        // Initialisation
        // ---------------------------------------------------------------

        /** @brief Cree les fenetres du viewer */
	void createWindows(int argc,
			   char*argv[],
			   bool map3D=true,
			   bool logPlayer=false);
	/** @brief Rafraichit l'ecran */
	void updateDisplay(Second currentTime);

	/** @brief Autorise ou non l'affiche du point correspondant a la 
            position du robot */
	void enableDisplayRobots(bool enable);

        /** @brief Definit la forme du robot, et sa couleur */
        void setRobotModel(int                  robotId,
                           std::string const&   name;
                           ViewerRobotModelEN   texture,
                           ViewerColorST const& color);
        /** @brief Defini la position des balles de Squatch */
        void setBridgePosition(BridgePosition pos);
        /** @brief Defini la position des plateaux fixes supportant les 
            quilles */
	void setSupportPosition(Point const& pt1, 
                                Point const& pt2);

        // ---------------------------------------------------------------
        // Mettre a jour la position des objets sur le terrains
        // ---------------------------------------------------------------
        
        /** @brief Definit la position de la camera autour de la table 
            pour la vue 3D */ 
	void setCameraPosition(Millimeter cx, 
			       Millimeter cy, 
			       Millimeter cz);
	/** @brief Met a jour la position du robot */
	void setRobotPosition(int robotId,
                              Millimeter x, 
                              Millimeter y, 
                              Radian theta); 
	/** @brief Met a jour la position du robot */
	void setRobotPosition(int robotId,
                              Position const& pos);
        /** @brief Defini la position de toutes les quilles sur le jeu */
        void setSkittlePosition(Skittle* skittles);
        /** @brief Defini la position des balles de GRS */
        void setGRSBallsPosition(GRSBall* balls);
        /** @brief Defini la position des balles de Squatch */
        void setSquatchBallsPosition(SquatchBall* balls);
        
        // ---------------------------------------------------------------
        // Donnees de debug concernant chaque robot
        // ---------------------------------------------------------------
        
        /** @brief Trajectoire cible du robot */
	void setRobotTrajectory(int robotId,
                                Trajectory const& t);
        /** @brief Met a jour la constante de temps concernant les zones sur
            lesquelles le robot est deja passe. La zone sur laquelle vient
            de passer le robot et coloriee puis la couleur s'estompe avec le
            temps */
	void setCoveredZoneEraseTime(Millisecond delay);
        /** @brief Met a jour la liste d'obstacles detectes par un robot */
	void setObstacleList(int                    robotId,
                             ListOfObstacles const& obstacles,
			     ObstacleType           type);
	/** @brief Met a jour la grille de calcul de trajectoire */
	/*void setGrid(int      robotId,
                     LogGrid* grid);*/
        
        // ---------------------------------------------------------------
        // Films et screenshot
        // ---------------------------------------------------------------
        
        /** @brief Sauvegarde une fenetre dans un fichier image bmp */
	void screenShot(const char* filename);
	/** @brief Commence l'enregistrement d'une serie d'images 
	    correspondant a une fenetre */
	void movieStart(const char* filename, 
                        ViewerScreenEN screen);
	void movieStop(ViewerScreenEN screen);
	/** @brief Renvoie le nom de fichier utilise pour sauver les images 
	    d'un film */

        // ---------------------------------------------------------------
        // Callbacks des boutons du menu
        // ---------------------------------------------------------------
        void registerCallBack(ViewerButton, FunctionPtr btnClicked);
        
 protected:
        /** @brief constructeur accessible uniquement par la fonction 
            instance() */
        Viewer3DCL();
	/** @brief Lance le thread qui initialise les fenetres OPENGL */
	void startThread();
        
        // ---------------------------------------------------------------
        // Dessin 3D
        // ---------------------------------------------------------------
        void draw3D();
	/** @brief Dessine le RS en 3D */
	void drawRobot3D(int robotId);
	/** @brief Dessine le terrain et la salle en 3D */
	void drawBG3D();
	/** @brief Dessine toutes les quilles */
	void drawSkittles3D();
	/** @brief Dessine une quille */
	void drawSkittle3D(Skittle* skittle);
	/** @brief Dessine toutes les balles */
	void drawGRSBalls3D();
	/** @brief Dessine une balle de GRS */
	void drawGRSBall3D(GRSBall* ball);
        /** @brief Dessine toutes les balles */
	void drawSquatchBalls3D();
	/** @brief Dessine une balle de Squatch */
	void drawSquatchBall3D(SquatchBall* ball);
        /** @brief Dessine tous les ponts */
	void drawBridges3D();
	/** @brief Dessine un pont */
	void drawBridge3D(Millimeter x, bool big);

        // ---------------------------------------------------------------
        // Dessin 2D
        // ---------------------------------------------------------------
        void draw2D();
	/** @brief Dessine le RS en 3D */
	void drawRobot2D(int robotId);
	/** @brief Dessine le terrain et la salle en 3D */
	void drawBG2D();
	/** @brief Dessine toutes les quilles */
	void drawSkittles2D();
	/** @brief Dessine une quille */
	void drawSkittle2D(Skittle* skittle);
	/** @brief Dessine toutes les balles */
	void drawGRSBalls2D();
	/** @brief Dessine une balle de GRS */
	void drawGRSBall2D(GRSBall* ball);
        /** @brief Dessine toutes les balles */
	void drawSquatchBalls2D();
	/** @brief Dessine une balle de Squatch */
	void drawSquatchBall2D(SquatchBall* ball);
        /** @brief Dessine tous les ponts */
	void drawBridges2D();
	/** @brief Dessine un pont */
	void drawBridge2D(Millimeter y, bool big);
        /** @brief Dessine tous les obstacles */
	void drawObstacles2D();
	/** @brief Dessine un obstacle */
	void drawObstacle2D(Obstacle const& obstacle);
        /** @brief Dessine tous les obstacles */
	void drawTrajectories2D();
	/** @brief Dessine un obstacle */
	void drawTrajectory2D(Trajectory const& t);

        // ---------------------------------------------------------------
        // Films et screen shot
        // ---------------------------------------------------------------
        /** @brief Sauvegarde une image faisant partie d'un film */
	void movieScreenShot(ViewerScreenEN screen);
	/** @brief Arrete l'enregistrement d'une serie d'images */
	int  getMovieBaseName(ViewerScreenEN screen, 
                              char*filename,
                              bool incCounter=true);

 protected:
	Position         robot_[VIEWER_MAX_ROBOT_NBR];
        std::string      robotName_[VIEWER_MAX_ROBOT_NBR];
	ListOfObstacles* obstacles_[VIEWER_MAX_ROBOT_NBR];
	Trajectory       trajectory_[VIEWER_MAX_ROBOT_NBR];
	Second           time_;
        Skittle*         skittles_;
        GRSBall*         grsBalls_;
        SquatchBall*     squatchBalls_;
        BridgePosition   bridge_;
        Point            support1_;
        Point            support2_;

        bool enableObjectDisplay_;
	bool useTexture_;
	bool createThread_;
	Texture* texture_;
	
 private:
	static Viewer3DCL* viewer_;
	MThreadId thread_;
 
 public:
	static MovieDataST movieData_[VIEWER_SCREEN_NBR];
};

#endif // __VIEWER_3D_H__
