/**
 * @file log.cpp
 *
 * @author Laurent Saint-Marcel
 *
 * Class qui permet d'envoyer des information de la telemetrie du robot sur
 * une socket UNIX vers un programme qui va enregistrer les donnes sur le
 * disque
 */

#define LOG_SOCKET_INFO
#include "log.h"

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>     /* For AFUNIX sockets */
#include <stdarg.h>

#ifndef LOG_BASIC
#include "robotPosition.h"
#include "robotTimer.h"
#include "movementManager.h"
#endif // LOG_BASIC

// ============================================================================
// ================================   class Log   =============================
// ============================================================================
Log* Log::log_=NULL;

bool Log::FLAG_SENDING=true;
bool Log::FLAG_PRINTF =true;

static const int LOG_MAX_RETRY =2;

// ---------------------------------------------------------------------------
// sigpipe_handler
// ---------------------------------------------------------------------------
extern "C" void sigpipe_handler(int sig) 
{
    printf("Client has been disconnected !\n");
}

// ---------------------------------------------------------------------------
// Log::Log
// ---------------------------------------------------------------------------
Log::Log(const RobotConfig* config): 
  RobotComponent("Log", CLASS_LOG, config), sock_(-1), 
  simuData_(), oldTime_(-1), simulated_(false)
{  
    setVerboseLevel(VERBOSE_DEBUG);
    if (log_) delete log_;
    log_=this;
    initSocket();
    reset();
}

// ---------------------------------------------------------------------------
// Log::~Log
// ---------------------------------------------------------------------------
Log::~Log()
{
    LOG_DEBUG("delete log\n");
    close();
    log_=NULL;
}

// ---------------------------------------------------------------------------
// Log::reset
// ---------------------------------------------------------------------------
bool Log::reset()
{
    LOG_DEBUG("reset\n");
    init_= FLAG_SENDING;
    return init_;
}

// ---------------------------------------------------------------------------
// Log::close
// ---------------------------------------------------------------------------
void Log::close() {
    if (sock_ > 0) {
        LOG_DEBUG("socket close\n");
        shutdown(sock_, SHUT_RDWR);
        ::close(sock_);
        sock_ = -1;
    }
    FLAG_SENDING = false;
}

// ---------------------------------------------------------------------------
// Log::validate
// ---------------------------------------------------------------------------
bool Log::validate()
{
  return false;
}

// ---------------------------------------------------------------------------
// Log::initSocket
// ---------------------------------------------------------------------------
// Cree la socket UNIX qui connecte le programme du robot au systeme de 
// sauvegarde de logs
// ---------------------------------------------------------------------------
void Log::initSocket() {
    /* Create socket on which to send. */
#ifdef LOG_DATAGRAM
    sock_ = socket(AF_UNIX, SOCK_DGRAM, 0);
#else
    sock_ = socket(AF_UNIX, SOCK_STREAM, 0);
#endif			
    if (sock_ < 0) {		
        LOG_DEBUG("Cannot open the socket\n");
        FLAG_SENDING = false;
        return;
    }
    
    /* Construct name of socket to send to. */
    name_.sun_family = AF_UNIX;
    strcpy(name_.sun_path, LOG_SOCKET_PATH);
    
    /* Send message. */
    LogPacketHeader packetHeader(LOG_TYPE_PING);
    sendPacket(packetHeader, NULL);
    if (!FLAG_SENDING) {
      LOG_DEBUG("Cannot connect to the socket %s\n", LOG_SOCKET_PATH);
    }
    signal(SIGPIPE, sigpipe_handler);
}

// ---------------------------------------------------------------------------
// Log::sendPacket
// ---------------------------------------------------------------------------
// Envoyer un packet compos d'un header et de donne sur la socket
// ---------------------------------------------------------------------------
void Log::sendPacket(LogPacketHeader& header, 
                     Byte* data) {
    if (FLAG_SENDING && sock_ >= 0) {
        int n = 0;
        int length=0;
        int retry=0;
        // send Header
        do {
            n = sendto(sock_, (&header)+length, 
                     sizeof(LogPacketHeader)-length, 0,
		     (struct sockaddr *)&name_, sizeof(struct sockaddr_un));
            if (n < 0) {
                LOG_DEBUG("Cannot send the packet\n");
                FLAG_SENDING = false;
                return;
            }
            length += n;
        } while (length < (int)sizeof(LogPacketHeader) 
		 && retry++<LOG_MAX_RETRY);
        if (retry >= LOG_MAX_RETRY) {
            LOG_DEBUG("Cannot send the packet\n");
            FLAG_SENDING = false;
            return;
        }

        // send Data
        if (header.length == 0 || data==NULL) return;
        length = 0;
        retry=0;
        do {
            n = sendto(sock_, data+length, header.length-length, 0,
		     (struct sockaddr *)&name_, sizeof(struct sockaddr_un));
            if (n < 0) {
                LOG_DEBUG("Cannot send the packet\n");
                FLAG_SENDING = false;
                return;
            }
            length+=n;
        } while(length < header.length && retry++<LOG_MAX_RETRY);
        if (retry >= LOG_MAX_RETRY) {
            LOG_DEBUG("Cannot send the packet\n");
            FLAG_SENDING = false;
            return;
        }
    }
}

// ---------------------------------------------------------------------------
// Log::message
// ---------------------------------------------------------------------------
// Enregistre un message de type texte
// ---------------------------------------------------------------------------
void Log::message(ClassId     classId, 
                  LogType     msgType,  
                  int         line, 
                  const char* file,
                  const char* msg, ...)
{	
    LogPacketHeader header(msgType);
    LogPacketMessage data;
    data.cid  = classId;
    data.line = line;

    strncpy(data.file, file, LOG_FILE_LENGTH);
    data.file[LOG_FILE_LENGTH-1]=0;

    va_list argp;
    va_start(argp, msg);
    vsnprintf(data.mesg, LOG_MESG_LENGTH, msg, argp);
    va_end(argp);
    data.mesg[LOG_MESG_LENGTH-1]=0;
    header.length=sizeof(LogPacketMessage)-LOG_MESG_LENGTH+strlen(data.mesg)+1;
    sendPacket(header, (Byte*)(&data));	
}

// ---------------------------------------------------------------------------
// Log::sendPosition
// ---------------------------------------------------------------------------
// Enregistre la position du robot
// ---------------------------------------------------------------------------
void Log::sendPosition()
{	
#ifndef LOG_BASIC
    LogPacketHeader header(LOG_TYPE_POSITION);
    LogPacketPosition data;
    data.x     = ROBOT_POS->x();
    data.y     = ROBOT_POS->y();
    data.theta = ROBOT_POS->thetaAbsolute();
    header.length=sizeof(LogPacketPosition);
    sendPacket(header, (Byte*)(&data));
#endif	
}

// ---------------------------------------------------------------------------
// Log::sendMotor
// ---------------------------------------------------------------------------
// Enregistre des donnees sur la consigne envoyee aux moteurs et la valeur
// des codeurs
// ---------------------------------------------------------------------------
void Log::sendMotor()
{
#ifndef LOG_BASIC	
    LogPacketHeader header(LOG_TYPE_MOTOR);
    LogPacketMotorInfo data;
    MVTMGR->getPWM(data.pwmLeft, data.pwmRight);
    MVTMGR->getCoderPosition(data.posLeft, data.posRight);
    header.length=sizeof(LogPacketPosition);
    sendPacket(header, (Byte*)(&data));
#endif	
}

// ---------------------------------------------------------------------------
// Log::sendButtons
// ---------------------------------------------------------------------------
// Enregistre la valeur de tous les capteurs de type boutons (non utilise)
// ---------------------------------------------------------------------------
void Log::sendButtons()
{
#ifndef LOG_BASIC	
    LogPacketHeader header(LOG_TYPE_BUTTONS);
    LogPacketbuttons data;
    header.length=sizeof(LogPacketbuttons);
    sendPacket(header, (Byte*)(&data));
#endif	
}


// ---------------------------------------------------------------------------
// Log::sendSimu
// ---------------------------------------------------------------------------
// Enregisre les donnees du simultateur (utilise qu'en mode simu)
// ---------------------------------------------------------------------------
void Log::sendSimu()
{
#ifndef LOG_BASIC
    if (simulated_) {	
        LogPacketHeader header(LOG_TYPE_SIMU_DATA);
	header.length=sizeof(LogPacketSimu);
	sendPacket(header, (Byte*)(&simuData_));
    }
#endif	
}
// ---------------------------------------------------------------------------
// Log::sendCPU
// ---------------------------------------------------------------------------
// Enregistre la charge du CPU (non implemente)
// ---------------------------------------------------------------------------
void Log::sendCPU()
{
#ifndef LOG_BASIC	
    LogPacketHeader header(LOG_TYPE_CPU);
    LogPacketCPU data;
    data.load=0;
    header.length=sizeof(LogPacketCPU);
    sendPacket(header, (Byte*)(&data));
#endif	
}

// ---------------------------------------------------------------------------
// Log::periodicTask
// ---------------------------------------------------------------------------
// Tache periodique qui enregistre les donnees qui doivent etre rafraichies 
// periodiquement
// ---------------------------------------------------------------------------
void Log::periodicTask(Millisecond time)
{
    if (abs(time-oldTime_)>100) {
        // n'envoie pas les donnees trop souvent...
#ifndef LOG_BASIC
        sendPosition();
        sendMotor();
        sendButtons();
        sendCPU();
        sendSimu();
#endif
	oldTime_=time;
    }
}

// ---------------------------------------------------------------------------
// logPeriodicTask
// ---------------------------------------------------------------------------
void logPeriodicTask(void* userData, Millisecond time)
{
    LOG->periodicTask(time);
}

// ---------------------------------------------------------------------------
// Log::enablePeriodicData
// ---------------------------------------------------------------------------
void Log::enablePeriodicData()
{
#ifndef LOG_BASIC
    TIMER->registerPeriodicFunction(logPeriodicTask, 
                                    "logPeriodicTask",
                                    this); 
#endif
}

// ---------------------------------------------------------------------------
// Log::startMatch
// ---------------------------------------------------------------------------
// Enregistre le signal de debut de match
// ---------------------------------------------------------------------------
void Log::startMatch()
{
    LogPacketHeader header(LOG_TYPE_START_MATCH);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::stopMatch
// ---------------------------------------------------------------------------
// Enregistre le signal de fin de match
// ---------------------------------------------------------------------------
void Log::stopMatch()
{
    LogPacketHeader header(LOG_TYPE_END_MATCH);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::newStrategy
// ---------------------------------------------------------------------------
// Enregistre le signal de l'execution d'une nouvelle strategie
// ---------------------------------------------------------------------------
void Log::newStrategy()
{
    LogPacketHeader header(LOG_TYPE_START_STRATEGY);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliArmLeft
// ---------------------------------------------------------------------------
// Enregistre l'etat du bras gauche
// ---------------------------------------------------------------------------
void Log::hliArmLeft(bool  armIn)
{
    LogPacketHeader header(armIn?LOG_TYPE_ARM_LEFT_IN:LOG_TYPE_ARM_LEFT_OUT);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliArmRight
// ---------------------------------------------------------------------------
// Enregistre l'etat du bras droit
// ---------------------------------------------------------------------------
void Log::hliArmRight(bool armIn)
{
    LogPacketHeader header(armIn?LOG_TYPE_ARM_RIGHT_IN:LOG_TYPE_ARM_RIGHT_OUT);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliLowerArm
// ---------------------------------------------------------------------------
void Log::hliLowerArm(bool armIn)
{
    LogPacketHeader header(armIn?LOG_TYPE_LOWER_ARM_IN:LOG_TYPE_LOWER_ARM_OUT);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliBackDoor
// ---------------------------------------------------------------------------
void Log::hliBackDoor(bool opened)
{
    LogPacketHeader header(opened?LOG_TYPE_BACK_DOOR_OPENED:
                           LOG_TYPE_BACK_DOOR_CLOSED);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliEntry
// ---------------------------------------------------------------------------
void Log::hliEntry(bool    open)
{
    LogPacketHeader header(open?LOG_TYPE_ENTRY_OPENED:LOG_TYPE_ENTRY_CLOSED);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliDrop
// ---------------------------------------------------------------------------
// Enregistre l'ordre de tir
// ---------------------------------------------------------------------------
void Log::hliDrop()
{
    LogPacketHeader header(LOG_TYPE_DROP_ONE_BALL);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::hliTouchDown
// ---------------------------------------------------------------------------
// Enregistre l'ordre de lacher les balles par l'arriere
// ---------------------------------------------------------------------------
void Log::hliTouchDown()
{
    LogPacketHeader header(LOG_TYPE_TOUCH_DOWN_ONE_BALL);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::sound
// ---------------------------------------------------------------------------
// Enregistre un son joue par le robot
// ---------------------------------------------------------------------------
void Log::sound(SoundId snd)
{
    LogPacketHeader header(LOG_TYPE_SOUND);
    LogPacketSound data(snd);
    header.length = sizeof(LogPacketSound);
    sendPacket(header, (Byte*)&data);
}

// ---------------------------------------------------------------------------
// Log::trajectory
// ---------------------------------------------------------------------------
// Enregistre les premiers points de la trajectoire du robot
// ---------------------------------------------------------------------------
void Log::trajectory(Trajectory const& t)
{
    LogPacketHeader header(LOG_TYPE_TRAJECTORY);
    LogPacketTrajectory data(t);
    header.length = sizeof(LogPacketTrajectory);
    sendPacket(header, (Byte*)&data);
}

// ---------------------------------------------------------------------------
// Log::obstacle
// ---------------------------------------------------------------------------
// Enregistre un obstacle
// ---------------------------------------------------------------------------
void Log::obstacle(Obstacle const& o,
		   int envDetectorId)
{
    LogPacketHeader header(LOG_TYPE_OBSTACLE);
    LogPacketObstacle data(o, envDetectorId);
    header.length = sizeof(LogPacketObstacle);
    sendPacket(header, (Byte*)&data);
}

// ---------------------------------------------------------------------------
// Log::ball
// ---------------------------------------------------------------------------
// Enregistre la detection d'une balle par les sharps
// ---------------------------------------------------------------------------
void Log::ball(Point const& pt,
	       int envDetectorId)
{
    LogPacketHeader header(LOG_TYPE_BALL);
    LogPacketBall data(pt, envDetectorId);
    header.length = sizeof(LogPacketBall);
    sendPacket(header, (Byte*)&data);
}

// ---------------------------------------------------------------------------
// Log::gonioBalise
// ---------------------------------------------------------------------------
// Enregistre la detection d'une balise
// ---------------------------------------------------------------------------
void Log::gonioBalise(GonioBalise baliseId,
		      Point const& gonioCenter,
		      Radian const& direction)
{
    LogPacketHeader header(LOG_TYPE_GONIO);
    LogPacketGonio data(baliseId, gonioCenter, direction);
    header.length = sizeof(LogPacketGonio);
    sendPacket(header, (Byte*)&data);
}

// ---------------------------------------------------------------------------
// Log::simuSetCamera
// ---------------------------------------------------------------------------
void Log::simuSetCamera(Millimeter x, Millimeter y, Millimeter z)
{
    LogPacketHeader header(LOG_TYPE_CAMERA);
    LogPacketCamera a(x,y,z);
    header.length = sizeof(LogPacketCamera);
    sendPacket(header, (Byte*)&a);
}

// ---------------------------------------------------------------------------
// Log::convertGrid
// ---------------------------------------------------------------------------
void Log::convertGrid(GridCentered<CaseData>& grid,
		      int score[HOMO_ROW_MAX],
		      LogGrid& logGrid)
{
    logGrid.sizeX=(char)grid.getXSize();
    logGrid.sizeY=(char)grid.getYSize();
    if (logGrid.elements==NULL)
	logGrid.elements = new char[logGrid.sizeX*logGrid.sizeY];
    int i=0,j=0,k=0;
    for(i=0;i<HOMO_ROW_MAX; i++) {
	logGrid.score[i]=(char)score[i];
    }
    char * v= logGrid.elements;
    for(i=0;i<logGrid.sizeX; i++) {
	for(j=0;j<logGrid.sizeY;j++) {
	    v[k++]=grid.getValue(i,j).type;
	}
    }
}
// ---------------------------------------------------------------------------
// Log::setGrid
// ---------------------------------------------------------------------------
// Enregistre la grille utilisee pour le calcul de trajectoire
// ---------------------------------------------------------------------------
void Log::setGrid(GridCentered<CaseData>& grid,
		  int score[HOMO_ROW_MAX])
{
    LogPacketHeader header(LOG_TYPE_GRID);
    static LogGrid logGrid;
    convertGrid(grid, score, logGrid);
    header.length = HOMO_ROW_MAX+2+logGrid.sizeX*logGrid.sizeY;
    sendPacket(header, (Byte*)&logGrid);

}

// ---------------------------------------------------------------------------
// Log::setCoconut
// ---------------------------------------------------------------------------
// Enregistre la position d'un cocotier en mode simu
// ---------------------------------------------------------------------------
void Log::setCoconut(Point coconut1)
{
    LogPacketHeader header(LOG_TYPE_COCONUT);
    LogPacketCoconut data(coconut1);
    header.length = sizeof(LogPacketCoconut);
    sendPacket(header, (Byte*)&data);
}

// ---------------------------------------------------------------------------
// Log::setSimuMode
// ---------------------------------------------------------------------------
// Enregister le mode dans lequel tourne le programme
// ---------------------------------------------------------------------------
void Log::setSimuMode(bool simu)
{
    LogPacketHeader header(simu?LOG_TYPE_SIMU:LOG_TYPE_REAL);
    sendPacket(header, NULL);
    simulated_=simu;
}

// ---------------------------------------------------------------------------
// Log::closeLogFile
// ---------------------------------------------------------------------------
// Ferme le log ce qui le squve sur le disque
// ---------------------------------------------------------------------------
void Log::closeLogFile()
{
    LogPacketHeader header(LOG_TYPE_CLOSE);
    sendPacket(header, NULL);
}

// ---------------------------------------------------------------------------
// Log::fillSimuStruct
// ---------------------------------------------------------------------------
void Log::fillSimuStruct(Position* posRS,
                         Position* posEnnemyRP,
                         Position* posEnnemyRS,
                         Ball*     balls)
{
    if (posRS)       simuData_.posRS       = *posRS;
    if (posEnnemyRP) simuData_.posEnnemyRP = *posEnnemyRP;
    if (posEnnemyRS) simuData_.posEnnemyRS = *posEnnemyRS;
    if (balls) {
      for(unsigned int i=0; i<BALLON_NBR; i++) {
	simuData_.ball[i] = balls[i];
      }
    }
}
