/*
 * Tools to generate trajectory in potential fields.
 * When you think of potential fields, picture in your mind either a charged
 * particle navigating through a magnetic field or a marble rolling down a 
 * hill. The basic idea is that behavior exhibited by the particle/marble will
 * depend on the combination of the shape of the field/hill. Unlike fields/hills
 * where the topology is externally specified by environmental conditions, the 
 * topology of the potential fields that a robot experiences are determined by
 * the designer. More specifically, the designer (a) creates multiple 
 * behaviors, each assigned a particular task or function, (b) represents
 * each of these behaviors as a potential field, and (c) combines all of the
 * behaviors to produce the robot's motion by combining the potential fields.
 * \n\n
 * The fundamental building block of potential elds is the action vector, 
 * which corresponds, roughly, to the speed and orientation of a moving robot.
 * Each behavior outputs a desired output vector. For example, consider a 
 * SeekGoal behavior that is assigned the task of making the robot head toward
 * an identied goal. The output of the SeekGoal behavior is a vector that
 * points the robot toward the goal.
 * \n\n
 * Ref: Potential Fields Tutorial - Michael A. Goodrich (Pfields.pdf)
 */

#include <stdlib.h> // random

#include "potentialFields.h"

// ---------------------------------------------------------------------------
// Private namespace: local functions
// ---------------------------------------------------------------------------
namespace {

    // -----------------------------------------------------------------------
    // isInRectangle
    // -----------------------------------------------------------------------
    // This function returns true if the point pt is in the rectangle
    // -----------------------------------------------------------------------
    inline bool isInRectangle(const Point lowerLeft, 
                              const Point upperRight, 
                              const Point pt) {
        return ((pt.x >= lowerLeft.x) && (pt.y >= lowerLeft.y) 
                && (pt.x <= upperRight.x) && (pt.y <= upperRight.y));
    }

    // -----------------------------------------------------------------------
    // isInCircle
    // -----------------------------------------------------------------------
    // This function returns true if the point pt is in the circle
    // -----------------------------------------------------------------------
    inline bool isInCircle(const Point      center, 
                           const Millimeter radius, 
                           const Point      pt) {
        return (dist(center, pt) <= radius);
    }

    // -----------------------------------------------------------------------
    // isInCone
    // -----------------------------------------------------------------------
    // This function returns true if the point pt is in the circle
    // -----------------------------------------------------------------------
    inline bool isInCone(const Point   center, 
                         const Radian  direction, 
                         const Radian  width, 
                         const Point   pt) {
        Radian theta = na2PI(dir(center, pt)-direction,
                                                     -M_PI);
        return fabs(theta) <= fabs(width);
    }
    
    // -----------------------------------------------------------------------
    // orderRectanglePoints
    // -----------------------------------------------------------------------
    // This function computes lowerLeft and upperRight of 2 points defining a 
    // rectangle
    // -----------------------------------------------------------------------
    inline void orderRectanglePoints(const Point corner1, const Point corner2,
                                     Point & lowerLeft, Point & upperRight) {
        lowerLeft.x  = min(corner1.x, corner2.x);
        lowerLeft.y  = min(corner1.y, corner2.y);
        upperRight.x = max(corner1.x, corner2.x);
        upperRight.y = max(corner1.y, corner2.y);
    }

} // Private namespace

// ---------------------------------------------------------------------------
// PtlVerticalLine::PtlVerticalLine
// ---------------------------------------------------------------------------
// A vertical line which fields vectors are orthogonal
// If the dist between the point and the line is less than 'width'
//   then Vector = infinite (invalide position)
// If the dist between the point and the line is greater than 'maxDist' 
//   then Vector=0
// Else the vector is orthogonal to the line and its norme is equal to 
//   constanteGain+proportionnalGain*dist.
//
// If the gains are positive, it is a repulsive line else it is an attractive
//   line
// ---------------------------------------------------------------------------
PtlVerticalLine::PtlVerticalLine(Millimeter x, 
                                 Millimeter width, 
                                 PtlGain    constanteGain, 
                                 PtlGain    proportionnalGain, 
                                 Millimeter maxDist)
    : PtlField(), x_(x), width_(width), constanteGain_(constanteGain),
      proportionnalGain_(proportionnalGain), maxDist_(maxDist)
{
    width_   = fabs(width);
    maxDist_ = fabs(maxDist);
}
 
// ---------------------------------------------------------------------------
// PtlVerticalLine::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlVerticalLine::getVector(Point pt) const
{
    Millimeter d=fabs(pt.x-x_);
    if (d < width_ 
        || (maxDist_ != PTL_INFINITE_DIST && d > maxDist_)) {
        return PTL_NULL_VECTOR;
    }
    PtlGain signe=sign(pt.x-x_);
    return PtlVector(signe*(constanteGain_+d*proportionnalGain_),
                     0);
    
}

// ---------------------------------------------------------------------------
// PtlVerticalLine::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlVerticalLine::isValid(Point pt) const
{
    if (constanteGain_ >= 0 
        && fabs(pt.x-x_) < width_) return false;
    else return true;
}

// ---------------------------------------------------------------------------
// PtlVerticalLine::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlVerticalLine::isEnd(Point pt) const
{
    if (constanteGain_ < 0 
        && fabs(pt.x-x_) <= width_) return true;
    else return false;
}





// ---------------------------------------------------------------------------
// PtlHorizontalLine::PtlHorizontalLine
// ---------------------------------------------------------------------------
// An horizontal line which fields vectors are orthogonal
// If the dist between the point and the line is less than 'width'
//   then Vector = infinite (invalide position)
// If the dist between the point and the line is greater than 'maxDist' 
//   then Vector=0
// Else the vector is orthogonal to the line and its norme is equal to 
//   constanteGain+proportionnalGain*dist
//
// If the gains are positive, it is a repulsive line else it is an attractive
//   line
// ---------------------------------------------------------------------------
PtlHorizontalLine::PtlHorizontalLine(Millimeter y, 
                                     Millimeter width, 
                                     PtlGain    constanteGain, 
                                     PtlGain    proportionnalGain, 
                                     Millimeter maxDist)
    : PtlField(), y_(y), width_(width), constanteGain_(constanteGain),
      proportionnalGain_(proportionnalGain), maxDist_(maxDist)
{
    width_   = fabs(width);
    maxDist_ = fabs(maxDist);
}
 
// ---------------------------------------------------------------------------
// PtlHorizontalLine::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlHorizontalLine::getVector(Point pt) const
{
    Millimeter d=fabs(pt.y-y_);
    if (d < width_ 
        || (maxDist_ != PTL_INFINITE_DIST && d > maxDist_)) {
        return PTL_NULL_VECTOR;
    }
    PtlGain signe=sign(pt.y-y_);
    return PtlVector(0,
                     signe*(constanteGain_+d*proportionnalGain_));
}

// ---------------------------------------------------------------------------
// PtlHorizontalLine::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlHorizontalLine::isValid(Point pt) const
{
    if (constanteGain_ >= 0 
        && fabs(pt.y-y_) < width_) return false;
    else return true;
}

// ---------------------------------------------------------------------------
// PtlHorizontalLine::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlHorizontalLine::isEnd(Point pt) const
{
    if (constanteGain_ < 0 
        && fabs(pt.y-y_) <= width_) return true;
    else return false;
}





// ---------------------------------------------------------------------------
// PtlCircle::PtlCircle
// ---------------------------------------------------------------------------
// A circle which fields vectors are in the radius direction
// If the dist between the point and the line is less than 'width'
//   then Vector = infinite (invalide position)
// If the dist between the point and the line is greater than 'maxDist' 
//   then Vector=0
// Else the vector is orthogonal to the line and its norme is equal to 
//   constanteGain+proportionnalGain*dist
//
// If the gains are positive, it is a repulsive line else it is an attractive
//   line
// ---------------------------------------------------------------------------
PtlCircle::PtlCircle(Point      center, 
                     Millimeter radius, 
                     PtlGain    constanteGain, 
                     PtlGain    proportionnalGain, 
                     Millimeter maxDist)
    : PtlField(), center_(center), radius_(radius), 
      constanteGain_(constanteGain), proportionnalGain_(proportionnalGain), 
      maxDist_(maxDist)
{
    radius_  = fabs(radius);
    maxDist_ = fabs(maxDist_);
}

// ---------------------------------------------------------------------------
// PtlCircle::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlCircle::getVector(Point pt) const
{
    Millimeter d=dist(center_, pt) - radius_;
    if (d < 0
        || (maxDist_ != PTL_INFINITE_DIST && d > maxDist_)) {
        return PTL_NULL_VECTOR;
    }
    Radian theta = dir(center_, pt);
    return PtlVector((constanteGain_+d*proportionnalGain_)*cos(theta),
                     (constanteGain_+d*proportionnalGain_)*sin(theta));
}

// ---------------------------------------------------------------------------
// PtlCircle::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlCircle::isValid(Point pt) const
{
    if (constanteGain_ >= 0 
        && isInCircle(center_, radius_, pt)) return false;
    else return true;
}

// ---------------------------------------------------------------------------
// PtlCircle::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlCircle::isEnd(Point pt) const
{
    if (constanteGain_ < 0 
        && isInCircle(center_, radius_, pt)) return true;
    else return false;
}




// ---------------------------------------------------------------------------
// PtlCircleTangent::PtlCircleTangent
// ---------------------------------------------------------------------------
// A circle which fields vectors are tangent (orthogonal to the radius)
// If the dist between the point and the line is less than 'width'
//   then Vector = infinite (invalide position)
// If the dist between the point and the line is greater than 'maxDist' 
//   then Vector=0
// Else the vector is orthogonal to the line and its norme is equal to 
//   constanteGain+proportionnalGain*dist
//
// If the gains are positive, the vector turns in counter clock wise else in 
//   clock wise
// ---------------------------------------------------------------------------
PtlCircleTangent::PtlCircleTangent(Point      center, 
                                   Millimeter radius, 
                                   PtlGain    constanteGain, 
                                   PtlGain    proportionnalGain, 
                                   Millimeter maxDist)
    : PtlField(), center_(center), radius_(radius), 
      constanteGain_(constanteGain), proportionnalGain_(proportionnalGain), 
      maxDist_(maxDist)
{
    radius_  = fabs(radius);
    maxDist_ = fabs(maxDist_);
}

// ---------------------------------------------------------------------------
// PtlCircleTangent::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlCircleTangent::getVector(Point pt) const
{
    Millimeter d=dist(center_, pt)-radius_;
    if (d < 0 
        || (maxDist_ != PTL_INFINITE_DIST && d > maxDist_)) {
        return PTL_NULL_VECTOR;
    }
    Radian theta = dir(center_, pt)+M_PI/2.;
    return PtlVector((constanteGain_+d*proportionnalGain_)*cos(theta),
                     (constanteGain_+d*proportionnalGain_)*sin(theta));
}

// ---------------------------------------------------------------------------
// PtlCircleTangent::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlCircleTangent::isValid(Point pt) const
{
    if (constanteGain_ >= 0 
        && isInCircle(center_, radius_, pt)) return false;
    else return true;
}

// ---------------------------------------------------------------------------
// PtlCircleTangent::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlCircleTangent::isEnd(Point pt) const
{
    if (constanteGain_ < 0 
        && isInCircle(center_, radius_, pt)) return true;
    else return false;
}




// ---------------------------------------------------------------------------
// PtlUniform
// ---------------------------------------------------------------------------
// An uniform field vector in the specified rectangle
// ---------------------------------------------------------------------------
PtlUniform::PtlUniform(Point      lowerLeft, 
                       Point      higherRight, 
                       PtlVector  vector)
    : PtlField(), lowerLeft_(), higherRight_(), 
      vector_(vector)
{
    orderRectanglePoints(lowerLeft, higherRight, lowerLeft_, higherRight_);
}

// ---------------------------------------------------------------------------
// PtlUniform::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlUniform::getVector(Point pt) const
{
    if (isInRectangle(lowerLeft_, higherRight_, pt)) return vector_;
    else return PTL_NULL_VECTOR;
}

// ---------------------------------------------------------------------------
// PtlCircleTangent::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlUniform::isValid(Point pt) const
{
    return true;
}

// ---------------------------------------------------------------------------
// PtlUniform::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlUniform::isEnd(Point pt) const
{
    return false;
}




// ---------------------------------------------------------------------------
// PtlRandom::PtlRandom
// ---------------------------------------------------------------------------
// A random field vector in the specified rectangle
// The norme of each vector is between minGain and maxGain 
// and the orientation between minDir and maxDir
// ---------------------------------------------------------------------------
PtlRandom::PtlRandom(Point      lowerLeft, 
                     Point      higherRight, 
                     PtlGain    minGain,
                     PtlGain    maxGain,
                     Radian     minDir,
                     Radian     maxDir)
    : PtlField(), lowerLeft_(), higherRight_(), 
      minGain_(0), deltaGain_(0), 
      minDir_(0), deltaDir_(0)
{
    minGain_   = min(minGain, maxGain);
    deltaGain_ = fabs(maxGain - minGain);
    minDir_   = min(minDir, maxDir);
    deltaDir_ = fabs(maxDir - minDir);
    orderRectanglePoints(lowerLeft, higherRight, lowerLeft_, higherRight_);
}

// ---------------------------------------------------------------------------
// PtlRandom::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlRandom::getVector(Point pt) const
{
    if (isInRectangle(lowerLeft_, higherRight_, pt)) {
        PtlGain gain  = minGain_ + (((double)rand())/RAND_MAX)*deltaGain_;
        Radian  theta = minDir_  + (((double)rand())/RAND_MAX)*deltaDir_;
        return PtlVector(gain*cos(theta), gain*sin(theta));
    } else {
        return PTL_NULL_VECTOR;
    }
}

// ---------------------------------------------------------------------------
// PtlRandom::isValid
// ---------------------------------------------------------------------------
// Return TRUE
// ---------------------------------------------------------------------------
bool PtlRandom::isValid(Point pt) const
{
    return true;
}

// ---------------------------------------------------------------------------
// PtlRandom::isEnd
// ---------------------------------------------------------------------------
// Return FALSE
// ---------------------------------------------------------------------------
bool PtlRandom::isEnd(Point pt) const
{
    return false;
}



// ---------------------------------------------------------------------------
// PtlCone::PtlCone
// ---------------------------------------------------------------------------
// A field vector in a cone.
// A cone is defined by a center, a direction and a width. All points 
//   pt / abs( dir(0, pt-center) -direction) <width are in the cone. 
// In the cone the potential field is orthogonal to the line center/direction
// If constantGain is positve, it is a repulsive field.
// ---------------------------------------------------------------------------
PtlCone::PtlCone(Point      center, 
                 Radian     direction,
                 Radian     width,
                 PtlGain    constantGain)
    : PtlField(), center_(center), direction_(direction), width_(width), 
      constantGain_(constantGain)
{
    width_ = fabs(width);
}

// ---------------------------------------------------------------------------
// PtlCone::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlCone::getVector(Point pt) const
{
    if (!isInCone(center_, direction_, width_, pt)) {
        return PTL_NULL_VECTOR;
    }
    PtlGain signe = 
        sign(na2PI(dir(center_, pt)-direction_,-M_PI));
    Radian theta = direction_ + signe*M_PI/2.;
    return PtlVector(constantGain_*cos(theta),
                     constantGain_*sin(theta));
    
}

// ---------------------------------------------------------------------------
// PtlCone::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlCone::isValid(Point pt) const
{
    return true;
}

// ---------------------------------------------------------------------------
// PtlCone::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlCone::isEnd(Point pt) const
{
    if (constantGain_ < 0 
        && isInCone(center_, direction_, width_, pt)) return true;
    else return false;
}



// ---------------------------------------------------------------------------
// PtlRectangle::PtlRectangle
// ---------------------------------------------------------------------------
// A Potential fields which vectors are orthogonal to a rectangle
// If the dist between the point and the rectangle border is less than 
//   'width' then Vector = infinite (invalide position)
// If the dist between the point and the rectangle border is greater than 
//   'maxDist' then Vector=0
// Else the vector is orthogonal to the line and its norme is equal to 
//   constanteGain+proportionnalGain*dist
//
// If the gains are positive, it is a repulsive line else it is an attractive
//   line
// ---------------------------------------------------------------------------
PtlRectangle::PtlRectangle(Point      lowerLeft, 
                           Point      higherRight,
                           PtlGain    constanteGain, 
                           PtlGain    proportionnalGain, 
                           Millimeter maxDist)
    : PtlField(), lowerLeft_(), higherRight_(), 
      constanteGain_(constanteGain), proportionnalGain_(proportionnalGain), 
      maxDist_(maxDist)
{
    maxDist_ = fabs(maxDist);
    orderRectanglePoints(lowerLeft, higherRight, lowerLeft_, higherRight_);
}

// ---------------------------------------------------------------------------
// PtlRectangle::getVector
// ---------------------------------------------------------------------------
// Return the vector associated to the specific point 
// ---------------------------------------------------------------------------
PtlVector PtlRectangle::getVector(Point pt) const
{
    if (pt.x > lowerLeft_.x && pt.x < higherRight_.x) {
        if (pt.y > lowerLeft_.y && pt.y < higherRight_.y) {
            return PTL_NULL_VECTOR;
        } else if (pt.y <= lowerLeft_.y) {
            Millimeter d = fabs(pt.y-lowerLeft_.y);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            return PtlVector(0, 
                             -(constanteGain_+proportionnalGain_*d));
        } else { // pt.y >= higherRight_.y
            Millimeter d = fabs(pt.y-higherRight_.y);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            return PtlVector(0, 
                             (constanteGain_+proportionnalGain_*d));
        }
    } else if (pt.x <= lowerLeft_.x) {
        if (pt.y > lowerLeft_.y && pt.y < higherRight_.y) {
            Millimeter d = fabs(pt.x-lowerLeft_.x);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            return PtlVector(-(constanteGain_+proportionnalGain_*d),
                             0);
        } else if (pt.y <= lowerLeft_.y) {
            Millimeter d = dist(lowerLeft_, pt);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            Radian theta = dir(lowerLeft_, pt);
            return PtlVector((constanteGain_+proportionnalGain_*d)*cos(theta),
                             (constanteGain_+proportionnalGain_*d)*sin(theta));
        } else { // pt.y >= higherRight_.y
            Point corner(lowerLeft_.x, higherRight_.y);
            Millimeter d = dist(corner, pt);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            Radian theta = dir(corner, pt);
            return PtlVector((constanteGain_+proportionnalGain_*d)*cos(theta),
                             (constanteGain_+proportionnalGain_*d)*sin(theta));
        }
    } else { // pt.x >= higherRight_.x
        if (pt.y > lowerLeft_.y && pt.y < higherRight_.y) {
            Millimeter d = fabs(pt.x-higherRight_.x);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            return PtlVector((constanteGain_+proportionnalGain_*d),
                             0);
        } else if (pt.y <= lowerLeft_.y) {
            Point corner(higherRight_.x, lowerLeft_.y);
            Millimeter d = dist(corner, pt);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            Radian theta = dir(corner, pt);
            return PtlVector((constanteGain_+proportionnalGain_*d)*cos(theta),
                             (constanteGain_+proportionnalGain_*d)*sin(theta));
        } else { // pt.y >= higherRight_.y
            Millimeter d = dist(higherRight_, pt);
            if (maxDist_ != PTL_INFINITE_DIST && fabs(d) > maxDist_) 
                return PTL_NULL_VECTOR;
            Radian theta = dir(higherRight_, pt);
            return PtlVector((constanteGain_+proportionnalGain_*d)*cos(theta),
                             (constanteGain_+proportionnalGain_*d)*sin(theta));
        }
    }
}

// ---------------------------------------------------------------------------
// PtlRectangle::isValid
// ---------------------------------------------------------------------------
// Return FALSE if the point is inside a repulsive figure
// ---------------------------------------------------------------------------
bool PtlRectangle::isValid(Point pt) const
{
    if (constanteGain_ >= 0 
        && isInRectangle(lowerLeft_, higherRight_, pt)) return false;
    else return true;
}

// ---------------------------------------------------------------------------
// PtlRectangle::isEnd
// ---------------------------------------------------------------------------
// Return TRUE if the point is inside an attractive figure
// ---------------------------------------------------------------------------
bool PtlRectangle::isEnd(Point pt) const
{
    if (constanteGain_ < 0 
        && isInRectangle(lowerLeft_, higherRight_, pt)) return true;
    else return false;
}





// ---------------------------------------------------------------------------
// PtlWorld::PtlWorld
// ---------------------------------------------------------------------------
// class that merge potential fields and extract trajectories
// ---------------------------------------------------------------------------
PtlWorld::PtlWorld()
    : listOfObjects_(NULL), stepCB_(NULL), stepCBUserData_(NULL)
{
    listOfObjects_ = new PtlFieldList;
}

// ---------------------------------------------------------------------------
// PtlWorld::~PtlWorld
// ---------------------------------------------------------------------------
PtlWorld::~PtlWorld()
{
    if (listOfObjects_) delete listOfObjects_;
}

// ---------------------------------------------------------------------------
// PtlWorld::addObject
// ---------------------------------------------------------------------------
// Add a potential field to this world
// ---------------------------------------------------------------------------
void PtlWorld::addObject(PtlField* object)
{
    if (!listOfObjects_ || !object) return;
    listOfObjects_->push_back(object);
}

// ---------------------------------------------------------------------------
// PtlWorld::getTrajectory
// ---------------------------------------------------------------------------
// get a trajectory from startingPoint to an attractive point
// specfied in the list of objects. Return FALSE if no trajectory 
// reaches an attractive point
// ---------------------------------------------------------------------------
PtlError PtlWorld::getTrajectory(Point       startingPoint,
                                 Trajectory& trajectory)
{
    if (!listOfObjects_ || listOfObjects_->empty()) {
        return PTL_ERR_NULL_FIELD;
    }
    trajectory.push_back(startingPoint);
    Point latestPoint = startingPoint;
    while (!isEnd(latestPoint)) {
        if (!isValid(latestPoint)) {
            return PTL_ERR_INVALID_PT;
        }
        if (stepCB_) {
            if (!stepCB_(stepCBUserData_, this, latestPoint, trajectory)) {
                return PTL_ERR_USER_STOP;
            }
        }
        PtlVector vector = getVector(latestPoint);
        if (vector == PTL_NULL_VECTOR) {
            return PTL_ERR_NULL_VEC;
        }
        latestPoint += vector;
        trajectory.push_back(latestPoint);
        if (trajectory.size()> PTL_TRAJECTORY_LIMIT) {
            return PTL_ERR_TRAJECTORY_LIMIT;
        }
    }
    return PTL_SUCCESS; // end point reached
}

// ---------------------------------------------------------------------------
// PtlWorld::clearObjectList
// ---------------------------------------------------------------------------
// Clear the list of potential fields of this world
// ---------------------------------------------------------------------------
void PtlWorld::clearObjectList(bool deletePtr)
{
    if (!listOfObjects_ || listOfObjects_->empty()) return;
    if (deletePtr) {
        PtlFieldList::iterator it;
        for (it=listOfObjects_->begin(); it != listOfObjects_->end(); it++) {
            if ((*it)) delete *it;
        }
    }
    listOfObjects_->clear();
}

// ---------------------------------------------------------------------------
// PtlWorld::eraseLatestObjects
// ---------------------------------------------------------------------------
// Remove nbrObjects form the list (begining with the latest objects)
// ---------------------------------------------------------------------------
void PtlWorld::eraseLatestObjects(int nbrObjects,
                                  bool deletePtr)
{
    if (!listOfObjects_ || listOfObjects_->empty() || nbrObjects <= 0) return;
    for (int i=0; i<nbrObjects && !listOfObjects_->empty(); i++) {
        if (deletePtr) {
            PtlField* pfield = listOfObjects_->back();
            if (pfield) delete pfield;
        }
        listOfObjects_->pop_back();
    }
}

// ---------------------------------------------------------------------------
// PtlWorld::getVector
// ---------------------------------------------------------------------------
PtlVector PtlWorld::getVector(Point currentPoint) const
{
    PtlVector vector = PTL_NULL_VECTOR;
    if (listOfObjects_ && !listOfObjects_->empty()) {
        PtlFieldList::const_iterator it;
        for (it=listOfObjects_->begin(); it != listOfObjects_->end(); it++) {
            if ((*it)) {
                vector += (*it)->getVector(currentPoint);
            }
        }
    }
    return vector;
}

// ---------------------------------------------------------------------------
// PtlWorld::isValid
// ---------------------------------------------------------------------------
bool PtlWorld::isValid(Point currentPoint) const
{
    if (listOfObjects_ && !listOfObjects_->empty()) {
        PtlFieldList::const_iterator it;
        for (it=listOfObjects_->begin(); it != listOfObjects_->end(); it++) {
            if ((*it) && !(*it)->isValid(currentPoint)) {
                return false;
            }
        }
    }
    return true;
}

// ---------------------------------------------------------------------------
// PtlWorld::isEnd
// ---------------------------------------------------------------------------
bool PtlWorld::isEnd(Point currentPoint) const
{
    if (listOfObjects_ && !listOfObjects_->empty()) {
        PtlFieldList::const_iterator it;
        for (it=listOfObjects_->begin(); it != listOfObjects_->end(); it++) {
            if ((*it) && (*it)->isEnd(currentPoint)) {
                return true;
            }
        }
        return false;
    } else {
        return true;
    }
}

void printf(PtlError err)
{
    switch(err) {
    case PTL_SUCCESS:
        printf("PTL_SUCCESS\n");
        break;
    case PTL_ERR_NULL_FIELD:
        printf("PTL_ERR_NULL_FIELD\n");
        break;
    case PTL_ERR_INVALID_PT:
        printf("PTL_ERR_INVALID_PT\n");
        break;
    case PTL_ERR_NULL_VEC:
        printf("PTL_ERR_NULL_VEC\n");
        break;
    case PTL_ERR_TRAJECTORY_LIMIT:
        printf("PTL_ERR_TRAJECTORY_LIMIT\n");
        break;
    case PTL_ERR_USER_STOP:
        printf("PTL_ERR_USER_STOP\n");
        break;
    }
}

// ---------------------------------------------------------------------------
// PtlWorld::registerTrajectoryStepCallback
// ---------------------------------------------------------------------------
// Callback run for each step of the trajectory generation
// This function is run after having found each point of the trajectory.
// If it returns false, the iteration stop and return PTL_ERR_USER_STOP
// The call back can modify the list of fields: it can run addObject, 
// or eraseLatestObjects
// Only one function can be register. To unregister it, function=NULL
// ---------------------------------------------------------------------------
void PtlWorld::registerTrajectoryStepCallback(PtlCBFn function, 
                                              void*   userData)
{
    stepCB_ = function;
    stepCBUserData_ = userData;
}

// ---------------------------------------------------------------------------
// PtlWorld::validate
// ---------------------------------------------------------------------------
// Test the potential fields classes. Return true if there is a bug, false if 
// there is no 'known' bugs dectected...
// ---------------------------------------------------------------------------
bool PtlWorld::validate() 
{
    test_begin();

    PtlWorld world;
    Point startPt;
    Trajectory trajectory;
    PtlError error;

    // === PtlVerticalLine ===
    world.addObject(new PtlVerticalLine(20, 2, -1, 0, PTL_INFINITE_DIST));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==19);
    test_condition(trajectory[18].x==18 && trajectory[18].y==0);
    world.clearObjectList(true);

    // === PtlHorizontalLine ===
    world.addObject(new PtlHorizontalLine(20, 2, -1, 0, PTL_INFINITE_DIST));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==19);
    test_condition(trajectory[18].y==18 && trajectory[18].x==0);
    world.clearObjectList(true);

    // === PTL_ERR_NULL_VEC ===
    world.addObject(new PtlHorizontalLine(20, 2, -1, 0, 5));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_NULL_VEC); 
    world.clearObjectList(true);

    // === circle ===
    world.addObject(new PtlCircle(Point(20,20), 2, -1, 0, PTL_INFINITE_DIST));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==28);
    test_condition(trajectory[27].y>=19.0 && trajectory[27].x>=19.0);
    test_condition(trajectory[27].y<19.2 && trajectory[27].x<19.2);
    world.clearObjectList(true);

    // === PTL_ERR_INVALID_PT ===
    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), 1, 0, PTL_INFINITE_DIST));
    startPt=Point(19,19); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_INVALID_PT); 
    world.clearObjectList(true);

    // === rectangle ===
    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==27);
    test_condition(trajectory[26].y>=18.3 && trajectory[26].x>=18.3);
    test_condition(trajectory[26].y<18.5 && trajectory[26].x<18.5);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(40,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==27);
    test_condition(trajectory[26].x>=21.5 && trajectory[26].y>=18.3);
    test_condition(trajectory[26].x<21.7 && trajectory[26].y<18.5);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(40,40); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==27);
    test_condition(trajectory[26].x>=21.5 && trajectory[26].y>=21.5);
    test_condition(trajectory[26].x<21.7 && trajectory[26].y<21.7);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(0,40); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==27);
    test_condition(trajectory[26].x>=18.3 && trajectory[26].y>=21.5);
    test_condition(trajectory[26].x<18.5 && trajectory[26].y<21.7);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(20,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==19);
    test_condition(trajectory[18].x==20 && trajectory[18].y==18);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(20,40); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==19);
    test_condition(trajectory[18].x==20 && trajectory[18].y==22);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(0,20); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==19);
    test_condition(trajectory[18].x==18 && trajectory[18].y==20);
    world.clearObjectList(true);

    world.addObject(new PtlRectangle(Point(18,18), Point(22,22), -1, 0, PTL_INFINITE_DIST));
    startPt=Point(40,20); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==19);
    test_condition(trajectory[18].x==22 && trajectory[18].y==20);
    world.clearObjectList(true); 

    // uniform
    world.addObject(new PtlUniform(Point(-5,5), Point(5,-5), PtlVector(-1,-2)));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_NULL_VEC); 
    test_condition(trajectory.size()==4);
    test_condition(trajectory[3].x==-3 && trajectory[3].y==-6);
    world.clearObjectList(true);

    world.addObject(new PtlUniform(Point(-5,5), Point(5,-5), PtlVector(-1,-1)));
    startPt=Point(6,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_NULL_VEC); 
    test_condition(trajectory.size()==1);
    world.clearObjectList(true);

    // cone
    world.addObject(new PtlCone(Point(-2,-2), M_PI/4, M_PI/4, 1));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_NULL_VEC); 
    test_condition(trajectory.size()==4);
    test_condition(trajectory[3].x>=-2.2&& trajectory[3].y>=2.0);
    test_condition(trajectory[3].x<-2.0 && trajectory[3].y<2.2);
    
    startPt=Point(-3,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_NULL_VEC); 

    world.clearObjectList(true);

    // === PTL_ERR_TRAJECTORY_LIMIT ===
    world.addObject(new PtlHorizontalLine(102, 2, -1, 0, PTL_INFINITE_DIST));
    world.addObject(new PtlCircle(Point(-10,40), 20, 3, 0, 10));
    world.addObject(new PtlCircle(Point(30,70), 25, 3, 0, 10));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_ERR_TRAJECTORY_LIMIT); 
    world.clearObjectList(true);
    
    // === real world ===
    world.addObject(new PtlHorizontalLine(102, 2, -1, 0, PTL_INFINITE_DIST));
    world.addObject(new PtlCircle(Point(-10,40), 20, 3, 0, 10));
    world.addObject(new PtlCircle(Point(30,70), 10, 3, 0, 10));
    startPt=Point(0,0); trajectory.clear();
    error = world.getTrajectory(startPt, trajectory);
    test_condition(error == PTL_SUCCESS); 
    test_condition(trajectory.size()==143);
    test_condition(trajectory[142].x>=8.2 && trajectory[142].y>=100);
    test_condition(trajectory[142].x<8.4 && trajectory[142].y<100.2);
    world.clearObjectList(true);


    test_end();
}

#ifdef TEST_MAIN

int main(int argc, char*argv[])
{
    PtlWorld::validate();
    return 1;
}

#endif // TEST_MAIN
