#include "robotTimer.h"
#include <time.h>
#include "mthread.h"
#include "log.h"

// ----------------------------------------------------------------------------
// private namespace
// ----------------------------------------------------------------------------

namespace {
  pthread_mutex_t repositoryLock3 = PTHREAD_MUTEX_INITIALIZER;
  extern "C" {
    struct timeval chronometerTic;
  }
  void no_op(...){};
}

void* RobotTimerThreadBody(void*);

// ----------------------------------------------------------------------------
// Static members
// ----------------------------------------------------------------------------
RobotTimer* RobotTimer::timer_=NULL;

// ----------------------------------------------------------------------------
// RobotTimer::RobotTimer
// ----------------------------------------------------------------------------
RobotTimer::RobotTimer(const RobotConfig* config) : 
  RobotComponent("Timer", CLASS_ROBOT_TIMER, config), time_(-1)
{
  assert(timer_==NULL);
  timer_=this;

  LOG_FUNCTION();
  no_op(repositoryLock3);
  tic(&chronometerTic);
  reset();
  init_= (MTHREAD_CREATE("Timer Thread",
			 &thread_, 
			 NULL, 
			 RobotTimerThreadBody,
			 NULL)==0);
  if (!init_) {
    LOG_ERROR("Can not run thread RobotTimer\n");
  } else {
    LOG_OK("Initialisation Done\n");
  }
}

// -------------------------------------------------------------------------
// RobotTimer::~RobotTimer
// -------------------------------------------------------------------------
RobotTimer::~RobotTimer()
{
  LOG_WARNING("Destructing RobotTimer\n");
  reset();
  MTHREAD_CANCEL(thread_);
  timer_=NULL;
}

// -------------------------------------------------------------------------
// RobotTimer::time
// -------------------------------------------------------------------------
// Retourne l'heure actuelle 
// -------------------------------------------------------------------------
Millisecond RobotTimer::time()
{
  if (instance()->time_>=0) {
      //Lock localLock(&repositoryLock3);
      instance()->time_ = tac(&chronometerTic)/1000;
  }
  return instance()->time_;
}

// -------------------------------------------------------------------------
// RobotTimer::startMatch
// -------------------------------------------------------------------------
// Met le compteur a 0 
// -------------------------------------------------------------------------
void RobotTimer::startMatch()
{
  tic(&chronometerTic);
  instance()->time_=0;
}

// -------------------------------------------------------------------------
// RobotTimer::reset
// -------------------------------------------------------------------------
// unregister toutes les fonctions timer et remet l'heure a 
//  ROBOT_TIMER_RESET_VALUE 
// -------------------------------------------------------------------------
bool RobotTimer::reset()
{
  LOG_FUNCTION();
  Lock localLock(&repositoryLock3, config());
  time_=-1;
  callbacks_.clear();
  periodicTasks_.clear();
  return true;
}

// -------------------------------------------------------------------------
// RobotTimer::validate
// -------------------------------------------------------------------------
// Test le module et reourne true en cas de succes 
// -------------------------------------------------------------------------
bool RobotTimer::validate()
{
  return false;
}

// -------------------------------------------------------------------------
// RobotTimer::registerTimerFunction
// -------------------------------------------------------------------------
// Enregistre une fonction qui sera executee quand time()==time 
// -------------------------------------------------------------------------
void RobotTimer::registerTimerFunction(TimerFunction fn, 
				       const char* fnName,
				       void* userData,
				       Millisecond time)
{
  assert(time_<0);
  TimerRegisterFunction param;
  param.fn=fn;
  strncpy(param.fnName, fnName, TIMER_FUNCTION_NAME_LENGTH);
  param.userData=userData;
  Lock localLock(&repositoryLock3, config());
  callbacks_.push_back(std::pair<Millisecond, TimerRegisterFunction>(time, param));
  LOG_INFO("Register Timer Callback %s\n", fnName);
}

// -------------------------------------------------------------------------
// RobotTimer::registerPeriodicFunction
// -------------------------------------------------------------------------
// Enregistre une fonction qui est executee regulierement 
// -------------------------------------------------------------------------
void RobotTimer::registerPeriodicFunction(TimerFunction fn, 
					  const char* fnName,
					  void* userData)
{
  assert(time_<0);
  TimerRegisterFunction param;
  param.fn=fn;
  strncpy(param.fnName, fnName, TIMER_FUNCTION_NAME_LENGTH);
  param.userData=userData;
  Lock localLock(&repositoryLock3, config());
  periodicTasks_.push_back(param);
  LOG_INFO("Register Periodic Function %s\n", fnName);
}

// -------------------------------------------------------------------------
// RobotTimer::periodicTask
// -------------------------------------------------------------------------
void RobotTimer::periodicTask()
{
    time();
    Lock localLock(&repositoryLock3, config());
    // periodic tasks
    std::deque< TimerRegisterFunction >::iterator it = periodicTasks_.begin();
    for(;it != periodicTasks_.end(); it++) {
	if ((*it).fn) {
	    //execute la fonction
	    (*it).fn((*it).userData, time_);
	}
    }
    // special events
    std::deque<std::pair< Millisecond, TimerRegisterFunction > >::iterator it2 
	= callbacks_.begin();
    for(;it2 != callbacks_.end(); it2++) {
	if ((*it2).first >= 0 && (*it2).first < time_) {
	    if ((*it2).second.fn) {
		// execute le callback
		LOG_DEBUG("Executing: %s time= %d\n", 
			 (*it2).second.fnName, (*it2).first);
		(*it2).second.fn((*it2).second.userData, time_);
	    }
	    // set negative execution time => it won't be run 2 times
	    (*it2).first = -1; 
	}
    }
}

// -------------------------------------------------------------------------
// RobotTimerThreadBody
// -------------------------------------------------------------------------
void* RobotTimerThreadBody(void*)
{
  int ancien_mode=0;
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &ancien_mode);
  while(1) { 
    pthread_testcancel();
    usleep(10000);
    RobotTimer::instance()->periodicTask();
  }
  return NULL;
}


// -------------------------------------------------------------------------
// RobotTimer::tic
// -------------------------------------------------------------------------
void RobotTimer::tic(struct timeval* tv_before )
{
  gettimeofday(tv_before, NULL);
}

// -------------------------------------------------------------------------
// RobotTimer::tac
// -------------------------------------------------------------------------
int RobotTimer::tac(struct timeval* tv_before)
{
  struct timeval tv_after;
  int result;
  gettimeofday(&tv_after, NULL);
  
  result = (tv_after.tv_sec-tv_before->tv_sec)*1000000 +
    (tv_after.tv_usec - tv_before->tv_usec);

  return result;
}
